""" :mod: RegisterReplica

    ==================

    .. module: RegisterReplica

    :synopsis: register replica handler

    RegisterReplica operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

from DIRAC import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class RegisterReplica(DMSRequestOperationsBase):
    """
    .. class:: RegisterReplica

    RegisterReplica operation handler

    :param self: self reference
    :param ~DIRAC.RequestManagementSystem.Client.Operation.Operation operation: Operation instance
    :param str csPath: CS path for this handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor"""
        DMSRequestOperationsBase.__init__(self, operation, csPath)

    def __call__(self):
        """call me maybe"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # RegisterReplica specific monitor info
            gMonitor.registerActivity(
                "RegisterReplicaAtt",
                "Attempted replicas registrations",
                "RequestExecutingAgent",
                "Replicas/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "RegisterReplicaOK",
                "Successful replicas registrations",
                "RequestExecutingAgent",
                "Replicas/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "RegisterReplicaFail",
                "Failed replicas registrations",
                "RequestExecutingAgent",
                "Replicas/min",
                gMonitor.OP_SUM,
            )

        # # counter for failed replicas

        failedReplicas = 0
        # # catalog to use
        catalogs = self.operation.Catalog
        if catalogs:
            catalogs = [cat.strip() for cat in catalogs.split(",")]
        # # get waiting files
        waitingFiles = self.getWaitingFilesList()

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(waitingFiles)))

        # # loop over files
        registerOperations = {}
        successReplicas = 0
        for opFile in waitingFiles:

            if not self.rmsMonitoring:
                gMonitor.addMark("RegisterReplicaAtt", 1)

            # # get LFN
            lfn = opFile.LFN
            # # and others
            targetSE = self.operation.targetSEList[0]
            replicaTuple = (lfn, opFile.PFN, targetSE)
            # # call ReplicaManager
            registerReplica = self.dm.registerReplica(replicaTuple, catalogs)
            # # check results
            if not registerReplica["OK"] or lfn in registerReplica["Value"]["Failed"]:
                # There have been some errors

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("RegisterReplicaFail", 1)
                #        self.dataLoggingClient().addFileRecord( lfn, "RegisterReplicaFail", ','.join( catalogs ) if catalogs else "all catalogs", "", "RegisterReplica" )

                reason = registerReplica.get(
                    "Message", registerReplica.get("Value", {}).get("Failed", {}).get(lfn, "Unknown")
                )
                errorStr = "failed to register LFN %s: %s" % (lfn, str(reason))
                # FIXME: this is incompatible with the change made in the DM that we
                # ignore failures if successful in at least one catalog
                if lfn in registerReplica.get("Value", {}).get("Successful", {}) and isinstance(reason, dict):
                    # As we managed, let's create a new operation for just the remaining registration
                    errorStr += " - adding registerReplica operations to request"
                    for failedCatalog in reason:
                        key = "%s/%s" % (targetSE, failedCatalog)
                        newOperation = self.getRegisterOperation(
                            opFile, targetSE, type="RegisterReplica", catalog=failedCatalog
                        )
                        if key not in registerOperations:
                            registerOperations[key] = newOperation
                        else:
                            registerOperations[key].addFile(newOperation[0])
                    opFile.Status = "Done"
                else:
                    opFile.Error = errorStr
                    catMaster = True
                    if isinstance(reason, dict):
                        from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

                        for failedCatalog in reason:
                            catMaster = catMaster and FileCatalog()._getCatalogConfigDetails(failedCatalog).get(
                                "Value", {}
                            ).get("Master", False)
                    # If one targets explicitly a catalog and it fails or if it fails on the master catalog
                    if (catalogs or catMaster) and (
                        "file does not exist" in opFile.Error.lower() or "no such file" in opFile.Error.lower()
                    ):
                        # Check if the file really exists in SE, if not, consider this file registration as Done
                        res = self.dm.getReplicaMetadata(lfn, targetSE)
                        notExist = bool("No such file" in res.get("Value", {}).get("Failed", {}).get(lfn, ""))
                        if not notExist:
                            opFile.Status = "Failed"
                        else:
                            opFile.Status = "Done"
                    if opFile.Status != "Done":
                        failedReplicas += 1
                self.log.warn(errorStr)

            else:
                # All is OK
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                else:
                    gMonitor.addMark("RegisterReplicaOK", 1)
                    successReplicas += 1
                    self.log.verbose(
                        "Replica %s has been registered at %s"
                        % (lfn, ",".join(catalogs) if catalogs else "all catalogs")
                    )

                opFile.Status = "Done"

        # # if we have new replications to take place, put them at the end
        if registerOperations:
            self.log.info("adding %d operations to the request" % len(registerOperations))
        for operation in registerOperations.values():
            self.operation._parent.addOperation(operation)

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        # # final check
        infoStr = ""
        if successReplicas:
            infoStr = "%d replicas successfully registered" % successReplicas
        if failedReplicas:
            infoStr += ", %d replicas failed to register" % failedReplicas
        self.log.info("All replicas processed", infoStr)
        if failedReplicas:
            self.operation.Error = "some replicas failed to register"
            return S_ERROR(self.operation.Error)

        return S_OK()



""" Module holding function(s) creating the pilot wrapper.

    This is a DIRAC-free module, so it could possibly be used also outside of DIRAC installations.

    The main client of this module is the SiteDirector, that invokes the functions here more or less like this::

        pilotFilesCompressedEncodedDict = getPilotFilesCompressedEncodedDict(pilotFiles)
        localPilot = pilotWrapperScript(pilotFilesCompressedEncodedDict,
                                        pilotOptions,
                                        pilotExecDir)
        _writePilotWrapperFile(localPilot=localPilot)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tempfile
import base64
import bz2

pilotWrapperContent = """#!/bin/bash
if command -v python &> /dev/null; then
  py='python'
elif command -v python3 &> /dev/null; then
  py='python3'
elif command -v python2 &> /dev/null; then
  py='python2'
fi
/usr/bin/env $py << EOF

# imports
from __future__ import print_function

import os
import stat
import tempfile
import sys
import shutil
import base64
import bz2
import logging
import time
import tarfile
import hashlib

# setting up the logging
formatter = logging.Formatter(fmt='%%(asctime)s UTC %%(levelname)-8s %%(message)s', datefmt='%%Y-%%m-%%d %%H:%%M:%%S')
logging.Formatter.converter = time.gmtime
try:
  screen_handler = logging.StreamHandler(stream=sys.stdout)
except TypeError:  # python2.6
  screen_handler = logging.StreamHandler(strm=sys.stdout)
screen_handler.setFormatter(formatter)
logger = logging.getLogger('pilotLogger')
logger.setLevel(logging.DEBUG)
logger.addHandler(screen_handler)

# just logging the environment as first thing
logger.debug('===========================================================')
logger.debug('Environment of execution host\\n')
for key, val in os.environ.items():
  logger.debug(key + '=' + val)
logger.debug('===========================================================\\n')

# putting ourselves in the right directory
pilotExecDir = '%(pilotExecDir)s'
if not pilotExecDir:
  pilotExecDir = os.getcwd()
pilotWorkingDirectory = tempfile.mkdtemp(suffix='pilot', prefix='DIRAC_', dir=pilotExecDir)
pilotWorkingDirectory = os.path.realpath(pilotWorkingDirectory)
os.chdir(pilotWorkingDirectory)
logger.info("Launching dirac-pilot script from %%s" %%os.getcwd())
"""


def pilotWrapperScript(
    pilotFilesCompressedEncodedDict=None, pilotOptions="", pilotExecDir="", envVariables=None, location=""
):
    """Returns the content of the pilot wrapper script.

     The pilot wrapper script is a bash script that invokes the system python. Linux only.

    :param pilotFilesCompressedEncodedDict: this is a possible dict of name:compressed+encoded content files.
                       the proxy can be part of this, and of course the pilot files
    :type pilotFilesCompressedEncodedDict: dict
    :param pilotOptions: options with which to start the pilot
    :type pilotOptions: string
    :param pilotExecDir: pilot execution directory
    :type pilotExecDir: string
    :param envVariables: dictionary of environment variables
    :type envVariables: dict
    :param location: location where to get the pilot files
    :type location: string

    :returns: content of the pilot wrapper
    :rtype: string
    """

    if pilotFilesCompressedEncodedDict is None:
        pilotFilesCompressedEncodedDict = {}

    if envVariables is None:
        envVariables = {}

    compressedString = ""
    # are there some pilot files to unpack? Then we create the unpacking string
    for pfName, encodedPf in pilotFilesCompressedEncodedDict.items():
        compressedString += """
try:
  with open('%(pfName)s', 'wb') as fd:
    if sys.version_info < (3,):
      fd.write(bz2.decompress(base64.b64decode(\"\"\"%(encodedPf)s\"\"\")))
    else:
      fd.write(bz2.decompress(base64.b64decode(b'%(encodedPf)s')))
  os.chmod('%(pfName)s', stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
except Exception as x:
  print(x, file=sys.stderr)
  logger.error(x)
  shutil.rmtree(pilotWorkingDirectory)
  sys.exit(3)
""" % {
            "encodedPf": encodedPf.decode() if hasattr(encodedPf, "decode") else encodedPf,
            "pfName": pfName,
        }

    envVariablesString = ""
    for name, value in envVariables.items():  # are there some environment variables to add?
        envVariablesString += """
os.environ[\"%(name)s\"]=\"%(value)s\"
""" % {
            "name": name,
            "value": value,
        }

    # add X509_USER_PROXY to establish pilot env in Cluster WNs
    if "proxy" in pilotFilesCompressedEncodedDict:
        envVariablesString += """
os.environ['X509_USER_PROXY'] = os.path.join(pilotWorkingDirectory, 'proxy')
"""

    # now building the actual pilot wrapper

    localPilot = pilotWrapperContent % {"pilotExecDir": pilotExecDir}

    if compressedString:
        localPilot += (
            """
# unpacking lines
logger.info("But first unpacking pilot files")
%s
"""
            % compressedString
        )

    if envVariablesString:
        localPilot += (
            """
# Modifying the environment
%s
"""
            % envVariablesString
        )

    if location:
        localPilot += """
# Getting the pilot files
logger.info("Getting the pilot files from %(location)s")

location = '%(location)s'.replace(' ', '').split(',')

import random
random.shuffle(location)

# we try from the available locations
locs = [os.path.join('https://', loc) for loc in location]
locations = locs + [os.path.join(loc, 'pilot') for loc in locs]
# adding also, as last the cvmfs location dirac.egi.eu, but this won't contain a valid JSON
locations += ['file:/cvmfs/dirac.egi.eu/pilot/']

for loc in locations:
  print('Trying %%s' %% loc)

  # Getting the json, tar, and checksum file
  try:

    # urllib is different between python 2 and 3
    if sys.version_info < (3,):
      from urllib2 import urlopen as url_library_urlopen
      from urllib2 import URLError as url_library_URLError
    else:
      from urllib.request import urlopen as url_library_urlopen
      from urllib.error import URLError as url_library_URLError

    for fileName in ['pilot.json', 'pilot.tar', 'checksums.sha512']:
      # needs to distinguish whether urlopen method contains the 'context' param
      # in theory, it should be available from python 2.7.9
      # in practice, some prior versions may be composed of recent urllib version containing the param
      if 'context' in url_library_urlopen.__code__.co_varnames:
        import ssl
        context = ssl._create_unverified_context()
        remoteFile = url_library_urlopen(os.path.join(loc, fileName),
                                         timeout=10,
                                         context=context)

      else:
        remoteFile = url_library_urlopen(os.path.join(loc, fileName),
                                         timeout=10)

      localFile = open(fileName, 'wb')
      localFile.write(remoteFile.read())
      localFile.close()

      if fileName != 'pilot.tar':
        continue
      try:
        pt = tarfile.open('pilot.tar', 'r')
        pt.extractall()
        pt.close()
      except Exception as x:
        print("tarfile failed with message (this is normal!) %%s" %% repr(x), file=sys.stderr)
        logger.error("tarfile failed with message (this is normal!) %%s" %% repr(x))
        logger.warn("Trying tar command (tar -xvf pilot.tar)")
        res = os.system("tar -xvf pilot.tar")
        if res:
          logger.error("tar failed with exit code %%d, giving up (this is normal!)" %% int(res))
          print("tar failed with exit code %%d, giving up (this is normal!)" %% int(res), file=sys.stderr)
          raise
    # if we get here we break out of the loop of locations
    break
  except (url_library_URLError, Exception) as e:
    print('%%s unreacheable (this is normal!)' %% loc, file=sys.stderr)
    logger.error('%%s unreacheable (this is normal!)' %% loc)
    logger.exception(e)

else:
  print("None of the locations of the pilot files is reachable", file=sys.stderr)
  logger.error("None of the locations of the pilot files is reachable")
  sys.exit(-1)

# download was successful, now we check checksums
if os.path.exists('checksums.sha512'):
  checksumDict = {}
  chkSumFile = open('checksums.sha512', 'rt')
  for line in chkSumFile.read().split('\\n'):
    if not line.strip():  ## empty lines are ignored
      continue
    expectedHash, fileName = line.split('  ', 1)
    if not os.path.exists(fileName):
      continue
    logger.info('Checking %%r for checksum', fileName)
    fileHash = hashlib.sha512(open(fileName, 'rb').read()).hexdigest()
    if fileHash != expectedHash:
      print('Checksum mismatch for file %%r' %% fileName, file=sys.stderr)
      print('Expected %%r, found %%r' %%(expectedHash, fileHash), file=sys.stderr)
      logger.error('Checksum mismatch for file %%r', fileName)
      logger.error('Expected %%r, found %%r', expectedHash, fileHash)
      sys.exit(-1)
    logger.debug('Checksum matched')

""" % {
            "location": location
        }

    localPilot += (
        """
# now finally launching the pilot script (which should be called dirac-pilot.py)
cmd = "$py dirac-pilot.py %s"
logger.info('Executing: %%s' %% cmd)
sys.stdout.flush()
ret = os.system(cmd)

# and cleaning up
shutil.rmtree(pilotWorkingDirectory)

# did it fail?
if ret:
  sys.exit(1)

EOF
"""
        % pilotOptions
    )

    return localPilot


def getPilotFilesCompressedEncodedDict(pilotFiles, proxy=None):
    """this function will return the dictionary of pilot files names : encodedCompressedContent
     that we are going to send

    :param pilotFiles: list of pilot files (list of location on the disk)
    :type pilotFiles: list
    :param proxy: the proxy to send
    :type proxy: X509Chain
    """
    pilotFilesCompressedEncodedDict = {}

    for pf in pilotFiles:
        with open(pf, "r") as fd:
            pfContent = fd.read()
        pfContentEncoded = base64.b64encode(bz2.compress(pfContent.encode(), 9))
        pilotFilesCompressedEncodedDict[os.path.basename(pf)] = pfContentEncoded

    if proxy is not None:
        compressedAndEncodedProxy = base64.b64encode(bz2.compress(proxy.dumpAllToString()["Value"]))
        pilotFilesCompressedEncodedDict["proxy"] = compressedAndEncodedProxy

    return pilotFilesCompressedEncodedDict


def _writePilotWrapperFile(workingDirectory=None, localPilot=""):
    """write the localPilot string to a file, rurn the file name

    :param workingDirectory: the directory where to store the pilot wrapper file
    :type workingDirectory: string
    :param localPilot: content of the pilot wrapper
    :type localPilot: string

    :returns: file name of the pilot wrapper
    :rtype: string
    """

    fd, name = tempfile.mkstemp(suffix="_pilotwrapper.py", prefix="DIRAC_", dir=workingDirectory)
    with os.fdopen(fd, "w") as pilotWrapper:
        pilotWrapper.write(localPilot)
    return name



#!/usr/bin/env python
########################################################################
# File :   dirac-service
# Author : Adria Casajus
########################################################################
"""
This is a script to launch DIRAC services. Mostly internal.

Usage:
  dirac-service [options] ...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from DIRAC.ConfigurationSystem.Client.LocalConfiguration import LocalConfiguration
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.DISET.ServiceReactor import ServiceReactor
from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"


@Script()
def main():
    localCfg = Script.localCfg
    positionalArgs = localCfg.getPositionalArguments()
    if len(positionalArgs) == 0:
        gLogger.fatal("You must specify which server to run!")
        sys.exit(1)

    serverName = positionalArgs[0]
    localCfg.setConfigurationForServer(serverName)
    localCfg.addMandatoryEntry("Port")
    # localCfg.addMandatoryEntry( "HandlerPath" )
    localCfg.addMandatoryEntry("/DIRAC/Setup")
    localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
    localCfg.addDefaultEntry("LogLevel", "INFO")
    localCfg.addDefaultEntry("LogColor", True)
    resultDict = localCfg.loadUserData()
    if not resultDict["OK"]:
        gLogger.initialize(serverName, "/")
        gLogger.error("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    includeExtensionErrors()

    serverToLaunch = ServiceReactor()
    result = serverToLaunch.initialize(positionalArgs)
    if not result["OK"]:
        gLogger.error(result["Message"])
        sys.exit(1)

    result = serverToLaunch.serve()
    if not result["OK"]:
        gLogger.error(result["Message"])
        sys.exit(1)


if __name__ == "__main__":
    main()



""" policy that evaluates on how many tickets are open at the moment.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase

__RCSID__ = "$Id$"


class GGUSTicketsPolicy(PolicyBase):
    """
    The GGUSTicketsPolicy class is a policy class that evaluates on
    how many tickets are open at the moment.

    GGUSTicketsPolicy, given the number of GGUS tickets opened, proposes a new
    status for the element.
    """

    @staticmethod
    def _evaluate(commandResult):
        """
        Evaluate policy on opened tickets, using args (tuple).

        :returns:
            {
              'Status':Active|Probing,
              'Reason':'GGUSTickets: n unsolved',
            }
        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        # The command returns a list of dictionaries, with only one if thre is something,
        # otherwise an empty list.
        commandResult = commandResult[0]

        if "OpenTickets" not in commandResult:
            result["Status"] = "Error"
            result["Reason"] = "Expected OpenTickets key for GGUSTickets"
            return S_OK(result)

        openTickets = commandResult["OpenTickets"]

        if openTickets == 0:
            result["Status"] = "Active"
            result["Reason"] = "NO GGUSTickets unsolved"
        else:
            # Setting to Probing is way too aggresive, as we do not know the nature of the tickets
            result["Status"] = "Degraded"
            result["Reason"] = "%s GGUSTickets unsolved: %s" % (openTickets, commandResult["Tickets"])

        return S_OK(result)



"""
Handling the download of the shifter Proxy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
from DIRAC.ConfigurationSystem.Client.Helpers import Registry


def getShifterProxy(shifterType, fileName=False):
    """This method returns a shifter's proxy

    :param str shifterType: ProductionManager / DataManager...
    :param str fileName: file name

    :return: S_OK(dict)/S_ERROR()
    """
    if fileName:
        mkDir(os.path.dirname(fileName))
    opsHelper = Operations()
    userName = opsHelper.getValue(cfgPath("Shifter", shifterType, "User"), "")
    if not userName:
        return S_ERROR("No shifter User defined for %s" % shifterType)
    result = Registry.getDNForUsername(userName)
    if not result["OK"]:
        return result
    userDN = result["Value"][0]
    result = Registry.findDefaultGroupForDN(userDN)
    if not result["OK"]:
        return result
    defaultGroup = result["Value"]
    userGroup = opsHelper.getValue(cfgPath("Shifter", shifterType, "Group"), defaultGroup)
    vomsAttr = Registry.getVOMSAttributeForGroup(userGroup)
    if vomsAttr:
        gLogger.info("Getting VOMS [%s] proxy for shifter %s@%s (%s)" % (vomsAttr, userName, userGroup, userDN))
        result = gProxyManager.downloadVOMSProxyToFile(
            userDN, userGroup, filePath=fileName, requiredTimeLeft=86400, cacheTime=86400
        )
    else:
        gLogger.info("Getting proxy for shifter %s@%s (%s)" % (userName, userGroup, userDN))
        result = gProxyManager.downloadProxyToFile(
            userDN, userGroup, filePath=fileName, requiredTimeLeft=86400, cacheTime=86400
        )
    if not result["OK"]:
        return result
    chain = result["chain"]
    fileName = result["Value"]
    return S_OK({"DN": userDN, "username": userName, "group": userGroup, "chain": chain, "proxyFile": fileName})


def setupShifterProxyInEnv(shifterType, fileName=False):
    """Return the shifter's proxy and set it up as the default
    proxy via changing the environment.
    This method returns a shifter's proxy

    :param str shifterType: ProductionManager / DataManager...
    :param str fileName: file name

    :return: S_OK(dict)/S_ERROR()
    """
    result = getShifterProxy(shifterType, fileName)
    if not result["OK"]:
        return result
    proxyDict = result["Value"]
    os.environ["X509_USER_PROXY"] = proxyDict["proxyFile"]
    return result



#! /usr/bin/env python
"""
Clean the given directory or a list of directories by removing it and all the
contained files and subdirectories from the physical storage and from the
file catalogs.

Example:
  $ dirac-dms-clean-directory /formation/user/v/vhamar/newDir
  Cleaning directory /formation/user/v/vhamar/newDir ...  OK
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC import exit as DIRACExit, gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Name"))

    Script.parseCommandLine()

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        lfns = [lfn.strip().split()[0] for lfn in sorted(open(inputFileName, "r").read().splitlines())]
    else:
        lfns = [inputFileName]

    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    dm = DataManager()
    retVal = 0
    for lfn in [lfn for lfn in lfns if lfn]:
        gLogger.notice("Cleaning directory %r ... " % lfn)
        result = dm.cleanLogicalDirectory(lfn)
        if not result["OK"]:
            gLogger.error("Failed to clean directory", result["Message"])
            retVal = -1
        else:
            if not result["Value"]["Failed"]:
                gLogger.notice("OK")
            else:
                for folder, message in result["Value"]["Failed"].items():
                    gLogger.error("Failed to clean folder", "%r: %s" % (folder, message))
                    retVal = -1

        DIRACExit(retVal)


if __name__ == "__main__":
    main()



""" StorageManagerHandler is the implementation of the StorageManagementDB in the DISET framework """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import gLogger, S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.StorageManagementSystem.DB.StorageManagementDB import StorageManagementDB


class StorageManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB object"""

        cls.storageManagementDB = StorageManagementDB()
        return S_OK()

    ######################################################################
    #
    #  Example call back methods
    #

    types_updateTaskStatus = []

    @classmethod
    def export_updateTaskStatus(cls, sourceID, status, successful=[], failed=[]):
        """An example to show the usage of the callbacks."""
        gLogger.info("updateTaskStatus: Received callback information for ID %s" % sourceID)
        gLogger.info("updateTaskStatus: Status = '%s'" % status)
        if successful:
            gLogger.info("updateTaskStatus: %s files successfully staged" % len(successful))
            for lfn, time in successful:
                gLogger.info("updateTaskStatus: %s %s" % (lfn.ljust(100), time.ljust(10)))
        if failed:
            gLogger.info("updateTaskStatus: %s files failed to stage" % len(successful))
            for lfn, time in failed:
                gLogger.info("updateTaskStatus: %s %s" % (lfn.ljust(100), time.ljust(10)))
        return S_OK()

    ######################################################################
    #
    #  Monitoring methods
    #

    types_getTaskStatus = [six.integer_types]

    @classmethod
    def export_getTaskStatus(cls, taskID):
        """Obtain the status of the stage task from the DB."""
        res = cls.storageManagementDB.getTaskStatus(taskID)
        if not res["OK"]:
            gLogger.error("getTaskStatus: Failed to get task status", res["Message"])
        return res

    types_getTaskInfo = [six.integer_types]

    @classmethod
    def export_getTaskInfo(cls, taskID):
        """Obtain the metadata of the stage task from the DB."""
        res = cls.storageManagementDB.getTaskInfo(taskID)
        if not res["OK"]:
            gLogger.error("getTaskInfo: Failed to get task metadata", res["Message"])
        return res

    types_getTaskSummary = [six.integer_types]

    @classmethod
    def export_getTaskSummary(cls, taskID):
        """Obtain the summary of the stage task from the DB."""
        res = cls.storageManagementDB.getTaskSummary(taskID)
        if not res["OK"]:
            gLogger.error("getTaskSummary: Failed to get task summary", res["Message"])
        return res

    types_getTasks = [dict]

    @classmethod
    def export_getTasks(cls, condDict, older=None, newer=None, timeStamp="LastUpdate", orderAttribute=None, limit=None):
        """Get the replicas known to the DB."""
        res = cls.storageManagementDB.getTasks(
            condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, orderAttribute=orderAttribute, limit=limit
        )
        if not res["OK"]:
            gLogger.error("getTasks: Failed to get Cache replicas", res["Message"])
        return res

    types_removeStageRequests = [list]

    @classmethod
    def export_removeStageRequests(cls, replicaIDs):
        res = cls.storageManagementDB.removeStageRequests(replicaIDs)
        if not res["OK"]:
            gLogger.error("removeStageRequests: Failed to remove StageRequests", res["Message"])
        return res

    types_getCacheReplicas = [dict]

    @classmethod
    def export_getCacheReplicas(
        cls, condDict, older=None, newer=None, timeStamp="LastUpdate", orderAttribute=None, limit=None
    ):
        """Get the replcias known to the DB."""
        res = cls.storageManagementDB.getCacheReplicas(
            condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, orderAttribute=orderAttribute, limit=limit
        )
        if not res["OK"]:
            gLogger.error("getCacheReplicas: Failed to get Cache replicas", res["Message"])
        return res

    types_getStageRequests = [dict]

    @classmethod
    def export_getStageRequests(
        cls, condDict, older=None, newer=None, timeStamp="StageRequestSubmitTime", orderAttribute=None, limit=None
    ):
        """Get the replcias known to the DB."""
        res = cls.storageManagementDB.getStageRequests(
            condDict=condDict, older=older, newer=newer, timeStamp=timeStamp, orderAttribute=orderAttribute, limit=limit
        )
        if not res["OK"]:
            gLogger.error("getStageRequests: Failed to get Cache replicas", res["Message"])
        return res

    #
    #                                                Monitoring methods
    #
    ######################################################################

    ####################################################################
    #
    # setRequest is used to initially insert tasks and their associated files. Leaves files in New status.
    #

    types_setRequest = [dict, six.string_types, six.string_types, six.integer_types]

    @classmethod
    def export_setRequest(cls, lfnDict, source, callbackMethod, taskID):
        """This method allows stage requests to be set into the StagerDB"""
        res = cls.storageManagementDB.setRequest(lfnDict, source, callbackMethod, taskID)
        if not res["OK"]:
            gLogger.error("setRequest: Failed to set stage request", res["Message"])
        return res

    ####################################################################
    #
    # The state transition of Replicas method
    #

    types_updateReplicaStatus = [list, six.string_types]

    @classmethod
    def export_updateReplicaStatus(cls, replicaIDs, newReplicaStatus):
        """This allows to update the status of replicas"""
        res = cls.storageManagementDB.updateReplicaStatus(replicaIDs, newReplicaStatus)
        if not res["OK"]:
            gLogger.error("updateReplicaStatus: Failed to update replica status", res["Message"])
        return res

    ####################################################################
    #
    # The state transition of the Replicas from New->Waiting
    #

    types_updateReplicaInformation = [list]

    @classmethod
    def export_updateReplicaInformation(cls, replicaTuples):
        """This method sets the pfn and size for the supplied replicas"""
        res = cls.storageManagementDB.updateReplicaInformation(replicaTuples)
        if not res["OK"]:
            gLogger.error("updateRelicaInformation: Failed to update replica information", res["Message"])
        return res

    ####################################################################
    #
    # The state transition of the Replicas from Waiting->StageSubmitted
    #
    types_getStagedReplicas = []

    @classmethod
    def export_getStagedReplicas(cls):
        """This method obtains the replicas for which all replicas in the task are Staged/StageSubmitted"""
        res = cls.storageManagementDB.getStagedReplicas()
        if not res["OK"]:
            gLogger.error("getStagedReplicas: Failed to obtain Staged/StageSubmitted replicas", res["Message"])
        return res

    types_getWaitingReplicas = []

    @classmethod
    def export_getWaitingReplicas(cls):
        """This method obtains the replicas for which all replicas in the task are Waiting"""
        res = cls.storageManagementDB.getWaitingReplicas()
        if not res["OK"]:
            gLogger.error("getWaitingReplicas: Failed to obtain Waiting replicas", res["Message"])
        return res

    types_getOfflineReplicas = []

    @classmethod
    def export_getOfflineReplicas(cls):
        """This method obtains the replicas for which all replicas in the task are Offline"""
        res = cls.storageManagementDB.getOfflineReplicas()
        if not res["OK"]:
            gLogger.error("getOfflineReplicas: Failed to obtain Offline replicas", res["Message"])
        return res

    types_getSubmittedStagePins = []

    @classmethod
    def export_getSubmittedStagePins(cls):
        """This method obtains the number of files and size of the requests submitted for each storage element"""
        res = cls.storageManagementDB.getSubmittedStagePins()
        if not res["OK"]:
            gLogger.error("getSubmittedStagePins: Failed to obtain submitted request summary", res["Message"])
        return res

    types_insertStageRequest = [dict, [six.integer_types, six.integer_types]]

    @classmethod
    def export_insertStageRequest(cls, requestReplicas, pinLifetime):
        """This method inserts the stage request ID assocaited to supplied replicaIDs"""
        res = cls.storageManagementDB.insertStageRequest(requestReplicas, pinLifetime)
        if not res["OK"]:
            gLogger.error("insertStageRequest: Failed to insert stage request information", res["Message"])
        return res

    ####################################################################
    #
    # The state transition of the Replicas from StageSubmitted->Staged
    #

    types_setStageComplete = [list]

    @classmethod
    def export_setStageComplete(cls, replicaIDs):
        """This method updates the status of the stage request for the supplied replica IDs"""
        res = cls.storageManagementDB.setStageComplete(replicaIDs)
        if not res["OK"]:
            gLogger.error("setStageComplete: Failed to set StageRequest complete", res["Message"])
        return res

    ####################################################################
    #
    # The methods for finalization of tasks
    #
    # Daniela: useless method
    '''types_updateStageCompletingTasks = []
  def export_updateStageCompletingTasks(self):
    """ This method checks whether the file for Tasks in 'StageCompleting' status
    are all Staged and updates the Task status to Staged """
    res = cls.storageManagementDB.updateStageCompletingTasks()
    if not res['OK']:
      gLogger.error('updateStageCompletingTasks: Failed to update StageCompleting tasks.',res['Message'])
    return res
  '''

    types_setTasksDone = [list]

    @classmethod
    def export_setTasksDone(cls, taskIDs):
        """This method sets the status in the Tasks table to Done for the list of supplied task IDs"""
        res = cls.storageManagementDB.setTasksDone(taskIDs)
        if not res["OK"]:
            gLogger.error("setTasksDone: Failed to set status of tasks to Done", res["Message"])
        return res

    types_removeTasks = [list]

    @classmethod
    def export_removeTasks(cls, taskIDs):
        """This method removes the entries from TaskReplicas and Tasks with the supplied task IDs"""
        res = cls.storageManagementDB.removeTasks(taskIDs)
        if not res["OK"]:
            gLogger.error("removeTasks: Failed to remove Tasks", res["Message"])
        return res

    types_removeUnlinkedReplicas = []

    @classmethod
    def export_removeUnlinkedReplicas(cls):
        """This method removes Replicas which have no associated Tasks"""
        res = cls.storageManagementDB.removeUnlinkedReplicas()
        if not res["OK"]:
            gLogger.error("removeUnlinkedReplicas: Failed to remove unlinked Replicas", res["Message"])
        return res

    ####################################################################
    #
    # The state transition of the Replicas from *->Failed
    #

    types_updateReplicaFailure = [dict]

    @classmethod
    def export_updateReplicaFailure(cls, replicaFailures):
        """This method sets the status of the replica to failed with the supplied reason"""
        res = cls.storageManagementDB.updateReplicaFailure(replicaFailures)
        if not res["OK"]:
            gLogger.error("updateRelicaFailure: Failed to update replica failure information", res["Message"])
        return res

    ####################################################################
    #
    # Methods for obtaining Tasks, Replicas with supplied state
    #

    types_getTasksWithStatus = [six.string_types]

    @classmethod
    def export_getTasksWithStatus(cls, status):
        """This method allows to retrieve Tasks with the supplied status"""
        res = cls.storageManagementDB.getTasksWithStatus(status)
        if not res["OK"]:
            gLogger.error("getTasksWithStatus: Failed to get tasks with %s status" % status, res["Message"])
        return res

    types_getReplicasWithStatus = [six.string_types]

    @classmethod
    def export_getReplicasWithStatus(cls, status):
        """This method allows to retrieve replicas with the supplied status"""
        res = cls.storageManagementDB.getCacheReplicas({"Status": status})
        if not res["OK"]:
            gLogger.error("getReplicasWithStatus: Failed to get replicas with %s status" % status, res["Message"])
        return res

    types_getStageSubmittedReplicas = []

    @classmethod
    def export_getStageSubmittedReplicas(cls):
        """This method obtains the replica metadata and the stage requestID for the replicas in StageSubmitted status"""
        res = cls.storageManagementDB.getCacheReplicas({"Status": "StageSubmitted"})
        if not res["OK"]:
            gLogger.error("getStageSubmittedReplicas: Failed to obtain StageSubmitted replicas", res["Message"])
        return res

    types_wakeupOldRequests = [list, six.integer_types]

    @classmethod
    def export_wakeupOldRequests(cls, oldRequests, retryInterval):
        """get only StageRequests with StageRequestSubmitTime older than 1 day AND are still not staged
        delete these requests
        reset Replicas with corresponding ReplicaIDs to Status='New'
        """
        res = cls.storageManagementDB.wakeupOldRequests(oldRequests, retryInterval)
        if not res["OK"]:
            gLogger.error("wakeupOldRequests: Failed to wake up old requests", res["Message"])
        return res

    types_setOldTasksAsFailed = [six.integer_types]

    @classmethod
    def export_setOldTasksAsFailed(cls, daysOld):
        """
        Set Tasks older than "daysOld" number of days to Failed
        These tasks have already been retried every day for staging
        """
        res = cls.storageManagementDB.setOldTasksAsFailed(daysOld)
        if not res["OK"]:
            gLogger.error("setOldTasksAsFailed: Failed to set old Tasks to Failed state. ", res["Message"])
        return res

    types_getAssociatedReplicas = [list]

    @classmethod
    def export_getAssociatedReplicas(cls, replicaIDs):
        """
        Retrieve the list of Replicas that belong to the same Tasks as the provided list
        """
        res = cls.storageManagementDB.getAssociatedReplicas(replicaIDs)
        if not res["OK"]:
            gLogger.error("getAssociatedReplicas: Failed to get Associated Replicas. ", res["Message"])
        return res

    types_killTasksBySourceTaskID = [list]

    @classmethod
    def export_killTasksBySourceTaskID(cls, sourceTaskIDs):
        """Given SourceTaskIDs (jobIDs), this will cancel further staging of files for the corresponding tasks"""
        res = cls.storageManagementDB.killTasksBySourceTaskID(sourceTaskIDs)
        if not res["OK"]:
            gLogger.error("removeTasks: Failed to kill staging", res["Message"])
        return res

    types_getCacheReplicasSummary = []

    @classmethod
    def export_getCacheReplicasSummary(cls):
        """Reports breakdown of file number/size in different staging states across storage elements"""
        res = cls.storageManagementDB.getCacheReplicasSummary()
        if not res["OK"]:
            gLogger.error(" getCacheReplicasSummary: Failed to retrieve summary from server", res["Message"])
        return res



""" unit tests for Transformation Clients
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access,missing-docstring,invalid-name

import six
import unittest
import json
import mock

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.TransformationSystem.Client.TaskManager import TaskBase
from DIRAC.TransformationSystem.Client.RequestTasks import RequestTasks
from DIRAC.TransformationSystem.Client.Transformation import Transformation
from DIRAC.TransformationSystem.Client.Utilities import PluginUtilities
from DIRAC.TransformationSystem.Client.BodyPlugin.DummyBody import DummyBody


class reqValFake_C(object):
    def validate(self, opsInput):
        for ops in opsInput:
            if not len(ops):
                return {"OK": False}
            for f in ops:
                try:
                    if not f.LFN:
                        return {"OK": False}
                except Exception:
                    return {"OK": False}
        return {"OK": True}


reqValFake = reqValFake_C()


class ClientsTestCase(unittest.TestCase):
    """Base class for the clients test cases"""

    def setUp(self):
        self.mockTransClient = mock.MagicMock()
        self.mockTransClient.setTaskStatusAndWmsID.return_value = {"OK": True}

        self.mockReqClient = mock.MagicMock()

        self.taskBase = TaskBase(transClient=self.mockTransClient)
        self.pu = PluginUtilities(transClient=self.mockTransClient)

        self.requestTasks = RequestTasks(
            transClient=self.mockTransClient, requestClient=self.mockReqClient, requestValidator=reqValFake
        )
        self.transformation = Transformation()

        self.maxDiff = None

    def tearDown(self):
        pass


class TaskBaseSuccess(ClientsTestCase):
    def test_updateDBAfterTaskSubmission(self):
        res = self.taskBase.updateDBAfterTaskSubmission({})
        self.assertEqual(res["OK"], True)


class PluginUtilitiesSuccess(ClientsTestCase):
    def test_groupByReplicas(self):
        res = self.pu.groupByReplicas(
            {
                "/this/is/at.1": ["SE1"],
                "/this/is/at.12": ["SE1", "SE2"],
                "/this/is/at.2": ["SE2"],
                "/this/is/at_123": ["SE1", "SE2", "SE3"],
                "/this/is/at_23": ["SE2", "SE3"],
                "/this/is/at_4": ["SE4"],
            },
            "Flush",
        )
        self.assertTrue(res["OK"])
        self.assertEqual(
            res["Value"],
            [
                ("SE1", ["/this/is/at.1"]),
                ("SE1,SE2", ["/this/is/at.12"]),
                ("SE1,SE2,SE3", ["/this/is/at_123"]),
                ("SE2", ["/this/is/at.2"]),
                ("SE2,SE3", ["/this/is/at_23"]),
                ("SE4", ["/this/is/at_4"]),
            ],
        )

        res = self.pu.groupByReplicas(
            {
                "/this/is/at.123": ["SE1", "SE2", "SE3"],
                "/this/is/at.12": ["SE1", "SE2"],
                "/this/is/at.134": ["SE1", "SE3", "SE4"],
            },
            "Flush",
        )
        self.assertTrue(res["OK"])
        self.assertEqual(
            res["Value"],
            [
                ("SE1,SE2", ["/this/is/at.12"]),
                ("SE1,SE2,SE3", ["/this/is/at.123"]),
                ("SE1,SE3,SE4", ["/this/is/at.134"]),
            ],
        )


class RequestTasksSuccess(ClientsTestCase):
    def test_prepareTranformationTasks(self):

        # No tasks in input
        taskDict = {}
        res = self.requestTasks.prepareTransformationTasks("", taskDict, "owner", "ownerGroup", "/bih/boh/DN")
        self.assertTrue(res["OK"])
        self.assertEqual(len(taskDict), 0)

        # 3 tasks, 1 task not OK (in second transformation)
        taskDict = {123: {"TransformationID": 2, "TargetSE": "SE3", "b3": "bb3", "InputData": ""}}
        res = self.requestTasks.prepareTransformationTasks("", taskDict, "owner", "ownerGroup", "/bih/boh/DN")
        self.assertTrue(res["OK"])
        # We should "lose" one of the task in the preparation
        self.assertEqual(len(taskDict), 0)

        taskDict = {
            1: {
                "TransformationID": 1,
                "TargetSE": "SE1",
                "b1": "bb1",
                "Site": "MySite",
                "InputData": ["/this/is/a1.lfn", "/this/is/a2.lfn"],
            },
            2: {"TransformationID": 1, "TargetSE": "SE2", "b2": "bb2", "InputData": "/this/is/a1.lfn;/this/is/a2.lfn"},
            3: {"TransformationID": 2, "TargetSE": "SE3", "b3": "bb3", "InputData": ""},
        }

        res = self.requestTasks.prepareTransformationTasks("", taskDict, "owner", "ownerGroup", "/bih/boh/DN")
        self.assertTrue(res["OK"])
        # We should "lose" one of the task in the preparation
        self.assertEqual(len(taskDict), 2)
        for task in res["Value"].values():
            self.assertTrue(isinstance(task["TaskObject"], Request))
            self.assertEqual(task["TaskObject"][0].Type, "ReplicateAndRegister")
            try:
                self.assertEqual(task["TaskObject"][0][0].LFN, "/this/is/a1.lfn")
            except IndexError:
                self.assertEqual(task["TaskObject"][0].Status, "Waiting")
            try:
                self.assertEqual(task["TaskObject"][0][1].LFN, "/this/is/a2.lfn")
            except IndexError:
                self.assertEqual(task["TaskObject"][0].Status, "Waiting")

        # # test another (single) OperationType
        res = self.requestTasks.prepareTransformationTasks(
            "someType;LogUpload", taskDict, "owner", "ownerGroup", "/bih/boh/DN"
        )
        self.assertTrue(res["OK"])
        # We should "lose" one of the task in the preparation
        self.assertEqual(len(taskDict), 2)
        for task in res["Value"].values():
            self.assertTrue(isinstance(task["TaskObject"], Request))
            self.assertEqual(task["TaskObject"][0].Type, "LogUpload")

        # ## Multiple operations
        transBody = [
            ("ReplicateAndRegister", {"SourceSE": "FOO-SRM", "TargetSE": "BAR-SRM"}),
            ("RemoveReplica", {"TargetSE": "FOO-SRM"}),
        ]
        jsonBody = json.dumps(transBody)

        taskDict = {
            1: {
                "TransformationID": 1,
                "TargetSE": "SE1",
                "b1": "bb1",
                "Site": "MySite",
                "InputData": ["/this/is/a1.lfn", "/this/is/a2.lfn"],
            },
            2: {"TransformationID": 1, "TargetSE": "SE2", "b2": "bb2", "InputData": "/this/is/a1.lfn;/this/is/a2.lfn"},
            3: {"TransformationID": 2, "TargetSE": "SE3", "b3": "bb3", "InputData": ""},
        }

        res = self.requestTasks.prepareTransformationTasks(jsonBody, taskDict, "owner", "ownerGroup", "/bih/boh/DN")
        self.assertTrue(res["OK"])
        # We should "lose" one of the task in the preparation
        self.assertEqual(len(taskDict), 2)
        for task in res["Value"].values():
            self.assertTrue(isinstance(task["TaskObject"], Request))
            self.assertEqual(task["TaskObject"][0].Type, "ReplicateAndRegister")
            self.assertEqual(task["TaskObject"][1].Type, "RemoveReplica")
            try:
                self.assertEqual(task["TaskObject"][0][0].LFN, "/this/is/a1.lfn")
                self.assertEqual(task["TaskObject"][1][0].LFN, "/this/is/a1.lfn")
            except IndexError:
                self.assertEqual(task["TaskObject"][0].Status, "Waiting")
                self.assertEqual(task["TaskObject"][1].Status, "Waiting")
            try:
                self.assertEqual(task["TaskObject"][0][1].LFN, "/this/is/a2.lfn")
                self.assertEqual(task["TaskObject"][1][1].LFN, "/this/is/a2.lfn")
            except IndexError:
                self.assertEqual(task["TaskObject"][0].Status, "Waiting")
                self.assertEqual(task["TaskObject"][1].Status, "Waiting")

            self.assertEqual(task["TaskObject"][0].SourceSE, "FOO-SRM")
            self.assertEqual(task["TaskObject"][0].TargetSE, "BAR-SRM")
            self.assertEqual(task["TaskObject"][1].TargetSE, "FOO-SRM")


class TransformationSuccess(ClientsTestCase):
    def test_setGet(self):

        res = self.transformation.setTransformationName("TestTName")
        self.assertTrue(res["OK"])
        description = "Test transformation description"
        res = self.transformation.setDescription(description)
        longDescription = "Test transformation long description"
        res = self.transformation.setLongDescription(longDescription)
        self.assertTrue(res["OK"])
        res = self.transformation.setType("MCSimulation")
        self.assertTrue(res["OK"])
        res = self.transformation.setPlugin("aPlugin")
        self.assertTrue(res["OK"])

        # # Test DataOperation Body

        res = self.transformation.setBody("")
        self.assertTrue(res["OK"])
        self.assertEqual(self.transformation.paramValues["Body"], "")

        res = self.transformation.setBody("_requestType;RemoveReplica")
        self.assertTrue(res["OK"])
        self.assertEqual(self.transformation.paramValues["Body"], "_requestType;RemoveReplica")

        # #Json will turn tuples to lists and strings to unicode
        transBody = [
            [u"ReplicateAndRegister", {u"SourceSE": u"FOO-SRM", u"TargetSE": u"BAR-SRM"}],
            [u"RemoveReplica", {u"TargetSE": u"FOO-SRM"}],
        ]
        res = self.transformation.setBody(transBody)
        self.assertTrue(res["OK"])

        self.assertEqual(self.transformation.paramValues["Body"], json.dumps(transBody))

        # # This is not true if any of the keys or values are not strings, e.g., integers
        self.assertEqual(json.loads(self.transformation.paramValues["Body"]), transBody)

        with self.assertRaisesRegexp(TypeError, "Expected list"):
            self.transformation.setBody({"ReplicateAndRegister": {"foo": "bar"}})
        with self.assertRaisesRegexp(TypeError, "Expected tuple"):
            self.transformation.setBody(["ReplicateAndRegister", "RemoveReplica"])
        with self.assertRaisesRegexp(TypeError, "Expected 2-tuple"):
            self.transformation.setBody([("ReplicateAndRegister", "RemoveReplica", "LogUpload")])
        with self.assertRaisesRegexp(TypeError, "Expected string"):
            self.transformation.setBody([(123, "Parameter:Value")])
        with self.assertRaisesRegexp(TypeError, "Expected dictionary"):
            self.transformation.setBody([("ReplicateAndRegister", "parameter=foo")])
        with self.assertRaisesRegexp(TypeError, "Expected string"):
            self.transformation.setBody([("ReplicateAndRegister", {123: "foo"})])
        with self.assertRaisesRegexp(ValueError, "Unknown attribute"):
            self.transformation.setBody([("ReplicateAndRegister", {"Request": Request()})])
        with self.assertRaisesRegexp(TypeError, "Cannot encode"):
            self.transformation.setBody([("ReplicateAndRegister", {"Arguments": Request()})])

        # Check that all tuples are checked by passing first a valid one,
        # then a faulty one.
        # It is enough to check one case, unlike above
        with self.assertRaisesRegexp(TypeError, "Expected 2-tuple"):
            self.transformation.setBody([(u"RemoveReplica", {u"TargetSE": u"FOO-SRM"}), ("One", "too long", "tuple")])

        # Test setting a body plugin as body
        complexBody = DummyBody()
        self.transformation.setBody(complexBody)

    def test_SetGetReset(self):
        """Testing of the set, get and reset methods.

          set*()
          get*()
          setTargetSE()
          setSourceSE()
          getTargetSE()
          getSourceSE()
          reset()
        Ensures that after a reset all parameters are returned to their defaults
        """

        res = self.transformation.getParameters()
        self.assertTrue(res["OK"])
        defaultParams = res["Value"].copy()
        for parameterName, defaultValue in res["Value"].items():
            if isinstance(defaultValue, six.string_types):
                testValue = "TestValue"
            else:
                testValue = 99999
            # # set*

            setterName = "set%s" % parameterName
            self.assertTrue(hasattr(self.transformation, setterName))
            setter = getattr(self.transformation, setterName)
            self.assertTrue(callable(setter))
            res = setter(testValue)
            self.assertTrue(res["OK"])
            # # get*
            getterName = "get%s" % parameterName
            self.assertTrue(hasattr(self.transformation, getterName))
            getter = getattr(self.transformation, getterName)
            self.assertTrue(callable(getter))
            res = getter()
            self.assertTrue(res["OK"])
            self.assertTrue(res["Value"], testValue)

        res = self.transformation.reset()
        self.assertTrue(res["OK"])
        res = self.transformation.getParameters()
        self.assertTrue(res["OK"])
        for parameterName, resetValue in res["Value"].items():
            self.assertEqual(resetValue, defaultParams[parameterName])
        self.assertRaises(AttributeError, self.transformation.getTargetSE)
        self.assertRaises(AttributeError, self.transformation.getSourceSE)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TaskBaseSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PluginUtilitiesSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(RequestTasksSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TransformationSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
########################################################################
# :file:    dirac-dms-directory-sync
# :author:  Marko Petric
########################################################################
"""
Provides basic rsync functionality for DIRAC

Syncs the source destination folder recursivly into the target destination

If option --sync is used contend that is not in the source directory but is
only in the target directory will be deleted.

Example:

  e.g.: Download
    dirac-dms-directory-sync LFN Path
  or Upload
    dirac-dms-directory-sync Path LFN SE
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"


@Script()
def main():
    Script.registerSwitch("D", "sync", "Make target directory identical to source")
    Script.registerSwitch("j:", "parallel=", "Multithreaded download and upload")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "LFN:           Logical File Name (Path to directory)",
            "Path:          Local path to the file (Path to directory)",
        )
    )
    Script.registerArgument(
        (
            "Path:          Local path to the file (Path to directory)",
            "LFN:           Logical File Name (Path to directory)",
        )
    )
    Script.registerArgument(" SE:            DIRAC Storage Element", mandatory=False)
    Script.parseCommandLine(ignoreErrors=False)

    args = Script.getPositionalArgs()
    if len(args) > 3:
        Script.showHelp()

    sync = False
    parallel = 1
    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "s" or switch[0].lower() == "sync":
            sync = True
        if switch[0].lower() == "j" or switch[0].lower() == "parallel":
            parallel = int(switch[1])

    from DIRAC import S_OK, S_ERROR
    from DIRAC import gConfig, gLogger
    from DIRAC.Core.Utilities.List import breakListIntoChunks
    from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
    from DIRAC.Core.Utilities.File import mkDir
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC.Resources.Storage.StorageElement import StorageElement

    from multiprocessing import Manager

    listOfFailedFiles = Manager().list()

    def getSetOfLocalDirectoriesAndFiles(path):
        """Return a set of all directories and subdirectories and a set of
        files contained therein for a given local path
        """

        fullPath = os.path.abspath(path)

        if not os.path.isdir(fullPath):
            return S_ERROR("The path: " + fullPath + " does not exist!")

        directories = set()
        files = set()

        for dirname, dirnames, filenames in os.walk(path):
            # add path to all subdirectories first.
            for subdirname in dirnames:
                fullSubdirname = os.path.join(dirname, subdirname)
                fullSubdirname = os.path.abspath(fullSubdirname)
                fullSubdirname = fullSubdirname.replace(fullPath, "").lstrip("/")
                directories.add(fullSubdirname)
            # add path to all filenames.
            for filename in filenames:
                fullFilename = os.path.join(dirname, filename)
                fullFilename = os.path.abspath(fullFilename)
                fullFilename = fullFilename.replace(fullPath, "").lstrip("/")
                fileSize = os.path.getsize(fullPath + "/" + fullFilename)
                if fileSize > 0:
                    files.add((fullFilename, int(fileSize)))

        tree = {}
        tree["Directories"] = directories
        tree["Files"] = files

        return S_OK(tree)

    def getSetOfRemoteSubDirectoriesAndFiles(path, fc, directories, files):
        """
        Recursively traverses all the subdirectories of a directory and returns a set of directories and files
        """
        result = fc.listDirectory(path)
        if result["OK"]:
            if result["Value"]["Successful"]:
                for entry in result["Value"]["Successful"][path]["Files"]:
                    size = result["Value"]["Successful"][path]["Files"][entry]["MetaData"]["Size"]
                    files.add((entry, size))
                for entry in result["Value"]["Successful"][path]["SubDirs"]:
                    directories.add(entry)
                    res = getSetOfRemoteSubDirectoriesAndFiles(entry, fc, directories, files)
                    if not res["OK"]:
                        return S_ERROR("Error: " + res["Message"])
                return S_OK()
            else:
                return S_ERROR("Error: %s" % result["Value"])
        else:
            return S_ERROR("Error:" + result["Message"])

    def getSetOfRemoteDirectoriesAndFiles(fc, path):
        """
        Return a set of all directories and subdirectories and the therein contained files for a given LFN
        """
        directories = set()
        files = set()

        res = getSetOfRemoteSubDirectoriesAndFiles(path, fc, directories, files)
        if not res["OK"]:
            return S_ERROR("Could not list remote directory: " + res["Message"])

        return_directories = set()
        return_files = set()

        for myfile in files:
            return_files.add((myfile[0].replace(path, "").lstrip("/"), myfile[1]))

        for mydirectory in directories:
            return_directories.add(mydirectory.replace(path, "").lstrip("/"))

        tree = {}
        tree["Directories"] = return_directories
        tree["Files"] = return_files

        return S_OK(tree)

    def getContentToSync(upload, fc, source_dir, dest_dir):
        """
        Return list of files and directories to be create and deleted
        """

        if upload:
            res = getSetOfRemoteDirectoriesAndFiles(fc, dest_dir)
            if not res["OK"]:
                return S_ERROR(res["Message"])
            to_dirs = res["Value"]["Directories"]
            to_files = res["Value"]["Files"]

            res = getSetOfLocalDirectoriesAndFiles(source_dir)
            if not res["OK"]:
                return S_ERROR(res["Message"])
            from_dirs = res["Value"]["Directories"]
            from_files = res["Value"]["Files"]

        else:
            res = getSetOfLocalDirectoriesAndFiles(dest_dir)
            if not res["OK"]:
                return S_ERROR(res["Message"])
            to_dirs = res["Value"]["Directories"]
            to_files = res["Value"]["Files"]

            res = getSetOfRemoteDirectoriesAndFiles(fc, source_dir)
            if not res["OK"]:
                return S_ERROR(res["Message"])
            from_dirs = res["Value"]["Directories"]
            from_files = res["Value"]["Files"]

        # Create list of directories to delete
        dirs_delete = list(to_dirs - from_dirs)
        # Sort the list by depth of directory tree
        dirs_delete.sort(key=lambda s: -s.count("/"))
        # Create list of directories to create
        dirs_create = list(from_dirs - to_dirs)
        # Sort the list by depth of directory tree
        dirs_create.sort(key=lambda s: s.count("/"))

        # Flatten the list of pairs (filename, size) to list of filename
        files_delete = [pair[0] for pair in list(to_files - from_files)]
        files_create = [pair[0] for pair in list(from_files - to_files)]

        create = {}
        create["Directories"] = dirs_create
        create["Files"] = files_create

        delete = {}
        delete["Directories"] = dirs_delete
        delete["Files"] = files_delete

        tree = {}
        tree["Create"] = create
        tree["Delete"] = delete

        return S_OK(tree)

    def removeRemoteFiles(dm, lfns):
        """
        Remove file from the catalog
        """
        for lfnList in breakListIntoChunks(lfns, 100):
            res = dm.removeFile(lfnList)
            if not res["OK"]:
                return S_ERROR("Failed to remove files:" + lfnList + res["Message"])
            else:
                return S_OK()

    def removeStorageDirectoryFromSE(directory, storageElement):
        """
        Delete directory on selected storage element
        """

        se = StorageElement(storageElement, False)
        res = returnSingleResult(se.exists(directory))

        if not res["OK"]:
            return S_ERROR("Failed to obtain existence of directory" + res["Message"])

        exists = res["Value"]
        if not exists:
            return S_OK("The directory %s does not exist at %s " % (directory, storageElement))

        res = returnSingleResult(se.removeDirectory(directory, recursive=True))
        if not res["OK"]:
            return S_ERROR("Failed to remove storage directory" + res["Message"])

        return S_OK()

    def removeRemoteDirectory(fc, lfn):
        """
        Remove file from the catalog
        """
        storageElements = gConfig.getValue("Resources/StorageElementGroups/SE_Cleaning_List", [])

        for storageElement in sorted(storageElements):
            res = removeStorageDirectoryFromSE(lfn, storageElement)
            if not res["OK"]:
                return S_ERROR("Failed to clean storage directory at all SE:" + res["Message"])
        res = returnSingleResult(fc.removeDirectory(lfn, recursive=True))
        if not res["OK"]:
            return S_ERROR("Failed to clean storage directory at all SE:" + res["Message"])

        return S_OK("Successfully removed directory")

    def createLocalDirectory(directory):
        """
        Create local directory
        """
        mkDir(directory)
        if not os.path.exists(directory):
            return S_ERROR("Directory creation failed")
        return S_OK("Created directory successfully")

    def removeLocalFile(path):
        """
        Remove local file
        """
        try:
            os.remove(path)
        except OSError as e:
            return S_ERROR("File deletion failed:" + e.strerror)

        if os.path.isfile(path):
            return S_ERROR("File deleting failed")
        return S_OK("Removed file successfully")

    def removeLocaDirectory(path):
        """
        Remove local directory
        """
        try:
            os.rmdir(path)
        except OSError as e:
            return S_ERROR("Deleting directory failed: " + e.strerror)

        if os.path.isdir(path):
            return S_ERROR("Directory deleting failed")
        return S_OK("Removed directory successfully")

    def doUpload(fc, dm, result, source_dir, dest_dir, storage, delete, nthreads):
        """
        Wrapper for uploading files
        """
        if delete:
            lfns = [dest_dir + "/" + filename for filename in result["Value"]["Delete"]["Files"]]
            if len(lfns) > 0:
                res = removeRemoteFiles(dm, lfns)
                if not res["OK"]:
                    gLogger.fatal("Deleting of files: " + lfns + " -X- [FAILED]" + res["Message"])
                    DIRAC.exit(1)
                else:
                    gLogger.notice("Deleting " + ", ".join(lfns) + " -> [DONE]")

            for directoryname in result["Value"]["Delete"]["Directories"]:
                res = removeRemoteDirectory(fc, dest_dir + "/" + directoryname)
                if not res["OK"]:
                    gLogger.fatal("Deleting of directory: " + directoryname + " -X- [FAILED] " + res["Message"])
                    DIRAC.exit(1)
                else:
                    gLogger.notice("Deleting " + directoryname + " -> [DONE]")

        for directoryname in result["Value"]["Create"]["Directories"]:
            res = returnSingleResult(fc.createDirectory(dest_dir + "/" + directoryname))
            if not res["OK"]:
                gLogger.fatal("Creation of directory: " + directoryname + " -X- [FAILED] " + res["Message"])
                DIRAC.exit(1)
            else:
                gLogger.notice("Creating " + directoryname + " -> [DONE]")

        listOfFiles = result["Value"]["Create"]["Files"]
        # Check that we do not have too many threads
        if nthreads > len(listOfFiles):
            nthreads = len(listOfFiles)

        if nthreads == 0:
            return S_OK("Upload finished successfully")

        listOfListOfFiles = chunkList(listOfFiles, nthreads)
        res = runInParallel(
            arguments=[dm, source_dir, dest_dir, storage], listOfLists=listOfListOfFiles, function=uploadListOfFiles
        )
        if not res["OK"]:
            return S_ERROR("Upload of files failed")

        return S_OK("Upload finished successfully")

    def uploadListOfFiles(dm, source_dir, dest_dir, storage, listOfFiles, tID):
        """
        Wrapper for multithreaded uploading of a list of files
        """
        log = gLogger.getLocalSubLogger("[Thread %s] " % tID)
        threadLine = "[Thread %s]" % tID
        for filename in listOfFiles:
            destLFN = os.path.join(dest_dir, filename)
            res = returnSingleResult(dm.putAndRegister(destLFN, source_dir + "/" + filename, storage, None))
            if not res["OK"]:
                log.fatal(threadLine + " Uploading " + filename + " -X- [FAILED] " + res["Message"])
                listOfFailedFiles.append("%s: %s" % (destLFN, res["Message"]))
            else:
                log.notice(threadLine + " Uploading " + filename + " -> [DONE]")

    def doDownload(dm, result, source_dir, dest_dir, delete, nthreads):
        """
        Wrapper for downloading files
        """
        if delete:
            for filename in result["Value"]["Delete"]["Files"]:
                res = removeLocalFile(dest_dir + "/" + filename)
                if not res["OK"]:
                    gLogger.fatal("Deleting of file: " + filename + " -X- [FAILED] " + res["Message"])
                    DIRAC.exit(1)
                else:
                    gLogger.notice("Deleting " + filename + " -> [DONE]")

            for directoryname in result["Value"]["Delete"]["Directories"]:
                res = removeLocaDirectory(dest_dir + "/" + directoryname)
                if not res["OK"]:
                    gLogger.fatal("Deleting of directory: " + directoryname + " -X- [FAILED] " + res["Message"])
                    DIRAC.exit(1)
                else:
                    gLogger.notice("Deleting " + directoryname + " -> [DONE]")

        for directoryname in result["Value"]["Create"]["Directories"]:
            res = createLocalDirectory(dest_dir + "/" + directoryname)
            if not res["OK"]:
                gLogger.fatal("Creation of directory: " + directoryname + " -X- [FAILED] " + res["Message"])
                DIRAC.exit(1)
            else:
                gLogger.notice("Creating " + directoryname + " -> [DONE]")

        listOfFiles = result["Value"]["Create"]["Files"]
        # Chech that we do not have to many threads
        if nthreads > len(listOfFiles):
            nthreads = len(listOfFiles)

        if nthreads == 0:
            return S_OK("Upload finished successfully")

        listOfListOfFiles = chunkList(listOfFiles, nthreads)
        res = runInParallel(
            arguments=[dm, source_dir, dest_dir],
            listOfLists=listOfListOfFiles,
            function=downloadListOfFiles,
        )

        if not res["OK"]:
            return S_ERROR("Download of files failed")

        return S_OK("Upload finished successfully")

    def chunkList(alist, nchunks):
        """
        Split a list into a list of equaliy sized lists
        """
        avg = len(alist) / float(nchunks)
        out = []
        last = 0.0

        while last < len(alist):
            out.append(alist[int(last) : int(last + avg)])
            last += avg

        return out

    def downloadListOfFiles(dm, source_dir, dest_dir, listOfFiles, tID):
        """
        Wrapper for multithreaded downloading of a list of files
        """
        log = gLogger.getLocalSubLogger("[Thread %s] " % tID)
        threadLine = "[Thread %s]" % tID
        for filename in listOfFiles:
            sourceLFN = os.path.join(source_dir, filename)
            res = returnSingleResult(dm.getFile(sourceLFN, dest_dir + ("/" + filename).rsplit("/", 1)[0]))
            if not res["OK"]:
                log.fatal(threadLine + " Downloading " + filename + " -X- [FAILED] " + res["Message"])
                listOfFailedFiles.append("%s: %s" % (sourceLFN, res["Message"]))
            else:
                log.notice(threadLine + " Downloading " + filename + " -> [DONE]")

    def runInParallel(arguments, listOfLists, function):
        """
        Helper for execution of uploads and downloads in parallel
        """
        from multiprocessing import Process

        processes = []
        for tID, alist in enumerate(listOfLists):
            argums = arguments + [alist] + [tID]
            pro = Process(target=function, args=argums)
            pro.start()
            processes.append(pro)
        for process in processes:
            process.join()

        if any(process.exitcode == 1 for process in processes):
            return S_ERROR()
        return S_OK()

    def syncDestinations(upload, source_dir, dest_dir, storage, delete, nthreads):
        """
        Top level wrapper to execute functions
        """

        fc = FileCatalog()
        dm = DataManager()

        result = getContentToSync(upload, fc, source_dir, dest_dir)
        if not result["OK"]:
            return S_ERROR(result["Message"])

        if upload:
            res = doUpload(fc, dm, result, source_dir, dest_dir, storage, delete, nthreads)
            if not res["OK"]:
                return S_ERROR("Upload failed: " + res["Message"])
        else:
            res = doDownload(dm, result, source_dir, dest_dir, delete, nthreads)
            if not res["OK"]:
                return S_ERROR("Download failed: " + res["Message"])

        return S_OK("Mirroring successfully finished")

    def run(parameters, delete, nthreads):
        """
        The main user interface
        """

        source_dir = parameters[0]
        dest_dir = parameters[1]
        upload = False
        storage = None

        if len(parameters) == 3:
            storage = parameters[2]
            source_dir = os.path.abspath(source_dir)
            dest_dir = dest_dir.rstrip("/")
            upload = True
            if not os.path.isdir(source_dir):
                gLogger.fatal("Source directory does not exist")
                DIRAC.exit(1)

        if len(parameters) == 2:
            dest_dir = os.path.abspath(dest_dir)
            source_dir = source_dir.rstrip("/")
            if not os.path.isdir(dest_dir):
                gLogger.fatal("Destination directory does not exist")
                DIRAC.exit(1)

        res = syncDestinations(upload, source_dir, dest_dir, storage, delete, nthreads)
        if not res["OK"]:
            return S_ERROR(res["Message"])

        return S_OK("Successfully mirrored " + source_dir + " into " + dest_dir)

    # This is the execution
    returnValue = run(args, sync, parallel)
    if listOfFailedFiles:
        gLogger.error("Some file operations failed:\n\t", "\n\t".join(listOfFailedFiles))
        DIRAC.exit(1)
    if not returnValue["OK"]:
        gLogger.fatal(returnValue["Message"])
        DIRAC.exit(1)
    else:
        gLogger.notice(returnValue["Value"])
        DIRAC.exit(0)


if __name__ == "__main__":
    main()



import pytest

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR, convertToReturnValue, returnValueOrRaise


def test_Ok():
    retVal = S_OK("Hello world")
    assert retVal["OK"] is True
    assert retVal["Value"] == "Hello world"


def test_Error():
    retVal = S_ERROR("This is bad")
    assert retVal["OK"] is False
    assert retVal["Message"] == "This is bad"
    callStack = "".join(retVal["CallStack"])
    assert "Test_ReturnValues" in callStack
    assert "test_Error" in callStack


def test_ErrorWithCustomTraceback():
    retVal = S_ERROR("This is bad", callStack=["My callstack"])
    assert retVal["OK"] is False
    assert retVal["Message"] == "This is bad"
    assert retVal["CallStack"] == ["My callstack"]


class CustomException(Exception):
    pass


@convertToReturnValue
def _happyFunction():
    return {"12345": "Success"}


@convertToReturnValue
def _sadFunction():
    raise CustomException("I am sad")
    return {}


def test_convertToReturnValue():
    retVal = _happyFunction()
    assert retVal["OK"] is True
    assert retVal["Value"] == {"12345": "Success"}
    # Make sure exceptions are captured correctly
    retVal = _sadFunction()
    assert retVal["OK"] is False
    assert "CustomException" in retVal["Message"]
    # Make sure the exception is re-raised
    with pytest.raises(CustomException):
        returnValueOrRaise(_sadFunction())



""" ResourceStatus

Module that acts as a helper for knowing the status of a resource.
It takes care of switching between the CS and the RSS.
The status is kept in the RSSCache object, which is a small wrapper on top of DictCache

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from datetime import datetime, timedelta
import math
from time import sleep
import six

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Utilities.RSSCacheNoThread import RSSCache
from DIRAC.ResourceStatusSystem.Utilities.RssConfiguration import RssConfiguration
from DIRAC.ResourceStatusSystem.Utilities.InfoGetter import getPoliciesThatApply


@six.add_metaclass(DIRACSingleton)
class ResourceStatus(object):
    """
    ResourceStatus helper that connects to CS if RSS flag is not Active.
    It keeps the connection to the db / server as an object member, to avoid creating a new
    one massively.
    """

    def __init__(self, rssFlag=None):
        """
        Constructor, initializes the rssClient.
        """
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        self.rssConfig = RssConfiguration()
        self.__opHelper = Operations()
        self.rssClient = ResourceStatusClient()
        self.rssFlag = rssFlag
        if rssFlag is None:
            self.rssFlag = self.__getMode()

        cacheLifeTime = int(self.rssConfig.getConfigCache())

        # RSSCache only affects the calls directed to RSS, if using the CS it is not used.
        self.rssCache = RSSCache(cacheLifeTime, self.__updateRssCache)

    def getElementStatus(self, elementName, elementType, statusType=None, default=None, vO=None):
        """
        Helper function, tries to get information from the RSS for the given
        Element, otherwise, it gets it from the CS.

        :param elementName: name of the element or list of element names
        :type elementName: str, list
        :param elementType: type of the element (StorageElement, ComputingElement, FTS, Catalog)
        :type elementType: str
        :param statusType: type of the status (meaningful only when elementType==StorageElement)
        :type statusType: None, str, list
        :param default: defult value (meaningful only when rss is InActive)
        :type default: str
        :return: S_OK/S_ERROR
        :rtype: dict

        :Example:

        >>> getElementStatus('CE42', 'ComputingElement')
            S_OK( { 'CE42': { 'all': 'Active' } } } )
        >>> getElementStatus('SE1', 'StorageElement', 'ReadAccess')
            S_OK( { 'SE1': { 'ReadAccess': 'Banned' } } } )
        >>> getElementStatus('SE1', 'ThisIsAWrongElementType', 'ReadAccess')
            S_ERROR( xyz.. )
        >>> getElementStatus('ThisIsAWrongName', 'StorageElement', 'WriteAccess')
            S_ERROR( xyz.. )
        >>> getElementStatus('A_file_catalog', 'FileCatalog')
            S_OK( { 'A_file_catalog': { 'all': 'Active' } } } )
        >>> getElementStatus('SE1', 'StorageElement', ['ReadAccess', 'WriteAccess'])
            S_OK( { 'SE1': { 'ReadAccess': 'Banned' , 'WriteAccess': 'Active'} } } )
        >>> getElementStatus('SE1', 'StorageElement')
            S_OK( { 'SE1': { 'ReadAccess': 'Probing' ,
                             'WriteAccess': 'Active',
                             'CheckAccess': 'Degraded',
                             'RemoveAccess': 'Banned'} } } )
        >>> getElementStatus(['CE1', 'CE2'], 'ComputingElement')
            S_OK( {'CE1': {'all': 'Active'},
                   'CE2': {'all': 'Probing'}}}
        """

        allowedParameters = ["StorageElement", "ComputingElement", "FTS", "Catalog"]

        if elementType not in allowedParameters:
            return S_ERROR("%s in not in the list of the allowed parameters: %s" % (elementType, allowedParameters))

        # Apply defaults
        if not statusType:
            if elementType == "StorageElement":
                statusType = ["ReadAccess", "WriteAccess", "CheckAccess", "RemoveAccess"]
            elif elementType == "ComputingElement":
                statusType = ["all"]
            elif elementType == "FTS":
                statusType = ["all"]
            elif elementType == "Catalog":
                statusType = ["all"]

        if self.rssFlag:
            return self.__getRSSElementStatus(elementName, elementType, statusType, vO)
        else:
            return self.__getCSElementStatus(elementName, elementType, statusType, default)

    def setElementStatus(self, elementName, elementType, statusType, status, reason=None, tokenOwner=None):
        """Tries set information in RSS and in CS.

        :param elementName: name of the element
        :type elementName: str
        :param elementType: type of the element (StorageElement, ComputingElement, FTS, Catalog)
        :type elementType: str
        :param statusType: type of the status (meaningful only when elementType==StorageElement)
        :type statusType: str
        :param reason: reason for setting the status
        :type reason: str
        :param tokenOwner: owner of the token (meaningful only when rss is Active)
        :type tokenOwner: str
        :return: S_OK/S_ERROR
        :rtype: dict

        :Example:

        >>> setElementStatus('CE42', 'ComputingElement', 'all', 'Active')
            S_OK(  xyz.. )
        >>> setElementStatus('SE1', 'StorageElement', 'ReadAccess', 'Banned')
            S_OK(  xyz.. )
        """

        if self.rssFlag:
            return self.__setRSSElementStatus(elementName, elementType, statusType, status, reason, tokenOwner)
        else:
            return self.__setCSElementStatus(elementName, elementType, statusType, status)

    ################################################################################

    def __updateRssCache(self):
        """Method used to update the rssCache.

        It will try 5 times to contact the RSS before giving up
        """

        meta = {"columns": ["Name", "ElementType", "StatusType", "Status", "VO"]}

        for ti in range(5):
            rawCache = self.rssClient.selectStatusElement("Resource", "Status", meta=meta)
            if rawCache["OK"]:
                break
            self.log.warn("Can't get resource's status", rawCache["Message"] + "; trial %d" % ti)
            sleep(math.pow(ti, 2))
            self.rssClient = ResourceStatusClient()

        if not rawCache["OK"]:
            return rawCache
        return S_OK(getCacheDictFromRawData(rawCache["Value"]))

    ################################################################################

    def __getRSSElementStatus(self, elementName, elementType, statusType, vO):
        """Gets from the cache or the RSS the Elements status. The cache is a
            copy of the DB table. If it is not on the cache, most likely is not going
            to be on the DB.

            There is one exception: item just added to the CS, e.g. new Element.
            The period between it is added to the DB and the changes are propagated
            to the cache will be inconsistent, but not dangerous. Just wait <cacheLifeTime>
            minutes.

        :param elementName: name of the element or list of element names
        :type elementName: str, list
        :param elementType: type of the element (StorageElement, ComputingElement, FTS, Catalog)
        :type elementType: str
        :param statusType: type of the status (meaningful only when elementType==StorageElement,
                           otherwise it is 'all' or ['all'])
        :type statusType: str, list
        """

        cacheMatch = self.rssCache.match(elementName, elementType, statusType, vO)

        self.log.debug("__getRSSElementStatus")
        self.log.debug(cacheMatch)

        return cacheMatch

    def __getCSElementStatus(self, elementName, elementType, statusType, default):
        """Gets from the CS the Element status

        :param elementName: name of the element
        :type elementName: str
        :param elementType: type of the element (StorageElement, ComputingElement, FTS, Catalog)
        :type elementType: str
        :param statusType: type of the status (meaningful only when elementType==StorageElement)
        :type statusType: str, list
        :param default: defult value
        :type default: None, str
        """

        # DIRAC doesn't store the status of ComputingElements nor FTS in the CS, so here we can just return 'Active'
        if elementType in ("ComputingElement", "FTS"):
            return S_OK({elementName: {"all": "Active"}})

        # If we are here it is because elementType is either 'StorageElement' or 'Catalog'
        if elementType == "StorageElement":
            cs_path = "/Resources/StorageElements"
        elif elementType == "Catalog":
            cs_path = "/Resources/FileCatalogs"
            statusType = ["Status"]

        if not isinstance(elementName, list):
            elementName = [elementName]

        if not isinstance(statusType, list):
            statusType = [statusType]

        result = {}
        for element in elementName:

            for sType in statusType:
                # Look in standard location, 'Active' by default
                res = gConfig.getValue("%s/%s/%s" % (cs_path, element, sType), "Active")
                result.setdefault(element, {})[sType] = res

        if result:
            return S_OK(result)

        if default is not None:
            defList = [[el, statusType, default] for el in elementName]
            return S_OK(getDictFromList(defList))

        _msg = "Element '%s', with statusType '%s' is unknown for CS."
        return S_ERROR(DErrno.ERESUNK, _msg % (elementName, statusType))

    def __setRSSElementStatus(self, elementName, elementType, statusType, status, reason, tokenOwner):
        """
        Sets on the RSS the Elements status
        """

        expiration = datetime.utcnow() + timedelta(days=1)

        self.rssCache.acquireLock()
        try:
            res = self.rssClient.addOrModifyStatusElement(
                "Resource",
                "Status",
                name=elementName,
                elementType=elementType,
                status=status,
                statusType=statusType,
                reason=reason,
                tokenOwner=tokenOwner,
                tokenExpiration=expiration,
            )

            if res["OK"]:
                self.rssCache.refreshCache()

            if not res["OK"]:
                _msg = "Error updating Element (%s,%s,%s)" % (elementName, statusType, status)
                gLogger.warn("RSS: %s" % _msg)

            return res

        finally:
            # Release lock, no matter what.
            self.rssCache.releaseLock()

    def __setCSElementStatus(self, elementName, elementType, statusType, status):
        """
        Sets on the CS the Elements status
        """

        # DIRAC doesn't store the status of ComputingElements nor FTS in the CS, so here we can just do nothing
        if elementType in ("ComputingElement", "FTS"):
            return S_OK()

        # If we are here it is because elementType is either 'StorageElement' or 'Catalog'
        statuses = self.rssConfig.getConfigStatusType(elementType)
        if statusType not in statuses:
            gLogger.error("%s is not a valid statusType" % statusType)
            return S_ERROR("%s is not a valid statusType: %s" % (statusType, statuses))

        if elementType == "StorageElement":
            cs_path = "/Resources/StorageElements"
        elif elementType == "Catalog":
            cs_path = "/Resources/FileCatalogs"
            # FIXME: This a probably outdated location (new one is in /Operations/[]/Services/Catalogs)
            # but needs to be VO-aware
            statusType = "Status"

        csAPI = CSAPI()
        csAPI.setOption("%s/%s/%s/%s" % (cs_path, elementName, elementType, statusType), status)

        res = csAPI.commitChanges()
        if not res["OK"]:
            gLogger.warn("CS: %s" % res["Message"])

        return res

    def __getMode(self):
        """
        Gets flag defined (or not) on the RSSConfiguration.
        If defined as 'Active', we use RSS, if not, we use the CS when possible (and WMS for Sites).
        """

        res = self.rssConfig.getConfigState()

        if res == "Active":
            if self.rssClient is None:
                self.rssClient = ResourceStatusClient()
            return True

        self.rssClient = None
        return False

    def isStorageElementAlwaysBanned(self, seName, statusType):
        """Checks if the AlwaysBanned policy is applied to the SE given as parameter

        :param seName: string, name of the SE
        :param statusType: ReadAcces, WriteAccess, RemoveAccess, CheckAccess

        :returns: S_OK(True/False)
        """

        res = getPoliciesThatApply({"name": seName, "statusType": statusType})
        if not res["OK"]:
            self.log.error("isStorageElementAlwaysBanned: unable to get the information", res["Message"])
            return res

        isAlwaysBanned = "AlwaysBanned" in [policy["type"] for policy in res["Value"]]

        return S_OK(isAlwaysBanned)


################################################################################


def getDictFromList(fromList):
    """
    Auxiliary method that given a list returns a dictionary of dictionaries:
    { site1 : { statusType1 : st1, statusType2 : st2 }, ... }
    """

    res = {}
    for listElement in fromList:
        site, sType, status = listElement
        if site not in res:
            res[site] = {}
        res[site][sType] = status
    return res


def getCacheDictFromRawData(rawList):
    """
    Formats the raw data list, which we know it must have tuples of five elements.
    ( element1, element2, element3, elementt4, element5 ) into a dictionary of tuples with the format
    { ( element1, element2, element3, element5 ): element4 )}.
    The resulting dictionary will be the new Cache.

    It happens that element1 is elementName,
                    element2 is elementType,
                    element3 is statusType,
                    element4 is status.
                    element5 is vO

    :Parameters:
      **rawList** - `list`
        list of three element tuples [( element1, element2, element3, element4, element5 ),... ]

    :return: dict of the form { ( elementName, elementType, statusType, vO ) : status, ... }
    """

    res = {}
    for entry in rawList:
        res.update({(entry[0], entry[1], entry[2], entry[4]): entry[3]})

    return res


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""
:mod: StorageElementProxyHandler

.. module: StorageElementProxyHandler

  :synopsis: This is a service which represents a DISET proxy to the Storage Element component.

This is used to get and put files from a remote storage.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
from six.moves import socketserver
import threading
import socket
import random
import six

# from DIRAC
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Core.Utilities.Subprocess import pythonCall
from DIRAC.Core.Utilities.Os import getDiskSpace
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.DataManagementSystem.private.HttpStorageAccessHandler import HttpStorageAccessHandler
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.ConfigurationSystem.Client.Helpers import Registry

__RCSID__ = "$Id$"

# globals
BASE_PATH = ""
HTTP_FLAG = False
HTTP_PORT = 9180
HTTP_PATH = ""


def purgeCacheDirectory(path):
    """del recursively :path:"""
    shutil.rmtree(path)


gRegister = DictCache(purgeCacheDirectory)


def initializeStorageElementProxyHandler(serviceInfo):
    """handler initialisation"""

    global BASE_PATH, HTTP_FLAG, HTTP_PORT, HTTP_PATH
    cfgPath = serviceInfo["serviceSectionPath"]

    BASE_PATH = gConfig.getValue("%s/BasePath" % cfgPath, BASE_PATH)
    if not BASE_PATH:
        gLogger.error("Failed to get the base path")
        return S_ERROR("Failed to get the base path")

    BASE_PATH = os.path.abspath(BASE_PATH)
    gLogger.info("The base path obtained is %s. Checking its existence..." % BASE_PATH)
    mkDir(BASE_PATH)

    HTTP_FLAG = gConfig.getValue("%s/HttpAccess" % cfgPath, False)
    if HTTP_FLAG:
        HTTP_PATH = "%s/httpCache" % BASE_PATH
        HTTP_PATH = gConfig.getValue("%s/HttpCache" % cfgPath, HTTP_PATH)
        mkDir(HTTP_PATH)
        HTTP_PORT = gConfig.getValue("%s/HttpPort" % cfgPath, 9180)
        gLogger.info("Creating HTTP server thread, port:%d, path:%s" % (HTTP_PORT, HTTP_PATH))
        _httpThread = HttpThread(HTTP_PORT, HTTP_PATH)

    return S_OK()


class ThreadedSocketServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """bag dummy class to hold ThreadingMixIn and TCPServer"""

    pass


class HttpThread(threading.Thread):
    """
    .. class:: HttpThread

    Single daemon thread running HttpStorageAccessHandler.
    """

    def __init__(self, port, path):
        """c'tor"""
        self.port = port
        self.path = path
        threading.Thread.__init__(self)
        self.setDaemon(1)
        self.start()

    def run(self):
        """thread run"""
        global gRegister
        handler = HttpStorageAccessHandler
        handler.register = gRegister
        handler.basePath = self.path
        httpd = ThreadedSocketServer(("", self.port), handler)
        httpd.serve_forever()


class StorageElementProxyHandler(RequestHandler):
    """
    .. class:: StorageElementProxyHandler
    """

    types_callProxyMethod = [six.string_types, six.string_types, list, dict]

    def export_callProxyMethod(self, se, name, args, kargs):
        """A generic method to call methods of the Storage Element."""
        res = pythonCall(200, self.__proxyWrapper, se, name, args, kargs)
        if res["OK"]:
            return res["Value"]
        return res

    def __proxyWrapper(self, se, name, args, kargs):
        """The wrapper will obtain the client proxy and set it up in the environment.

        The required functionality is then executed and returned to the client.
        """
        res = self.__prepareSecurityDetails()
        if not res["OK"]:
            return res
        credDict = self.getRemoteCredentials()
        group = credDict["group"]
        vo = Registry.getVOForGroup(group)
        if not vo:
            return S_ERROR("Can not determine VO of the operation requester")
        storageElement = StorageElement(se, vo=vo)
        method = getattr(storageElement, name) if hasattr(storageElement, name) else None
        if not method:
            return S_ERROR("Method '%s' isn't implemented!" % name)
        if not callable(getattr(storageElement, name)):
            return S_ERROR("Attribute '%s' isn't a method!" % name)
        return method(*args, **kargs)

    types_uploadFile = [six.string_types, six.string_types]

    def export_uploadFile(self, se, pfn):
        """This method uploads a file present in the local cache to the specified storage element"""
        res = pythonCall(300, self.__uploadFile, se, pfn)
        if res["OK"]:
            return res["Value"]
        return res

    def __uploadFile(self, se, pfn):
        """proxied upload file"""
        res = self.__prepareSecurityDetails()
        if not res["OK"]:
            return res

        # Put file to the SE
        try:
            storageElement = StorageElement(se)
        except AttributeError as x:
            errStr = "__uploadFile: Exception while instantiating the Storage Element."
            gLogger.exception(errStr, se, str(x))
            return S_ERROR(errStr)
        putFileDir = "%s/putFile" % BASE_PATH
        localFileName = "%s/%s" % (putFileDir, os.path.basename(pfn))
        res = returnSingleResult(storageElement.putFile({pfn: localFileName}))
        if not res["OK"]:
            gLogger.error("prepareFile: Failed to put local file to storage.", res["Message"])
        # Clear the local cache
        try:
            gLogger.debug("Removing temporary file", localFileName)
            os.remove(localFileName)
        except Exception as x:
            gLogger.exception("Failed to remove local file", localFileName, x)
        return res

    types_prepareFile = [six.string_types, six.string_types]

    def export_prepareFile(self, se, pfn):
        """This method simply gets the file to the local storage area"""
        res = pythonCall(300, self.__prepareFile, se, pfn)
        if res["OK"]:
            return res["Value"]
        return res

    def __prepareFile(self, se, pfn):
        """proxied prepare file"""
        res = self.__prepareSecurityDetails()
        if not res["OK"]:
            return res

        # Clear the local cache
        getFileDir = "%s/getFile" % BASE_PATH
        if not os.path.exists(getFileDir):
            os.mkdir(getFileDir)

        # Get the file to the cache
        try:
            storageElement = StorageElement(se)
        except AttributeError as x:
            errStr = "prepareFile: Exception while instantiating the Storage Element."
            gLogger.exception(errStr, se, str(x))
            return S_ERROR(errStr)
        res = returnSingleResult(storageElement.getFile(pfn, localPath="%s/getFile" % BASE_PATH))
        if not res["OK"]:
            gLogger.error("prepareFile: Failed to get local copy of file.", res["Message"])
            return res
        return S_OK()

    types_prepareFileForHTTP = [list(six.string_types) + [list]]

    def export_prepareFileForHTTP(self, lfn):
        """This method simply gets the file to the local storage area using LFN"""

        # Do clean-up, should be a separate regular thread
        gRegister.purgeExpired()

        key = str(random.getrandbits(128))
        result = pythonCall(300, self.__prepareFileForHTTP, lfn, key)
        if result["OK"]:
            result = result["Value"]
            # pylint believes it is a tuple because it is the only possible return type
            # it finds in Subprocess.py
            if result["OK"]:  # pylint: disable=invalid-sequence-index
                if HTTP_FLAG:
                    host = socket.getfqdn()
                    url = "http://%s:%d/%s" % (host, HTTP_PORT, key)
                    gRegister.add(key, 1800, result["CachePath"])
                    result["HttpKey"] = key
                    result["HttpURL"] = url
                return result
            return result
        return result

    def __prepareFileForHTTP(self, lfn, key):
        """Prepare proxied file for HTTP"""
        global HTTP_PATH

        res = self.__prepareSecurityDetails()
        if not res["OK"]:
            return res

        # Clear the local cache
        getFileDir = "%s/%s" % (HTTP_PATH, key)
        mkDir(getFileDir)

        # Get the file to the cache
        from DIRAC.DataManagementSystem.Client.DataManager import DataManager

        dataMgr = DataManager()
        result = dataMgr.getFile(lfn, destinationDir=getFileDir)
        result["CachePath"] = getFileDir
        return result

    ############################################################
    #
    # This is the method to setup the proxy and configure the environment with the client credential
    #

    def __prepareSecurityDetails(self):
        """Obtains the connection details for the client"""
        try:
            credDict = self.getRemoteCredentials()
            clientDN = credDict["DN"]
            clientUsername = credDict["username"]
            clientGroup = credDict["group"]
            gLogger.debug("Getting proxy for %s@%s (%s)" % (clientUsername, clientGroup, clientDN))
            res = gProxyManager.downloadVOMSProxy(clientDN, clientGroup)
            if not res["OK"]:
                return res
            chain = res["Value"]
            proxyBase = "%s/proxies" % BASE_PATH
            mkDir(proxyBase)
            proxyLocation = "%s/proxies/%s-%s" % (BASE_PATH, clientUsername, clientGroup)
            gLogger.debug("Obtained proxy chain, dumping to %s." % proxyLocation)
            res = gProxyManager.dumpProxyToFile(chain, proxyLocation)
            if not res["OK"]:
                return res
            gLogger.debug("Updating environment.")
            os.environ["X509_USER_PROXY"] = res["Value"]
            return res
        except Exception as error:
            exStr = "__getConnectionDetails: Failed to get client connection details."
            gLogger.exception(exStr, "", error)
            return S_ERROR(exStr)

    ############################################################
    #
    # These are the methods that are for actual file transfer
    #

    def transfer_toClient(self, fileID, token, fileHelper):
        """Method to send files to clients.
        fileID is the local file name in the SE.
        token is used for access rights confirmation.
        """
        file_path = "%s/%s" % (BASE_PATH, fileID)
        result = fileHelper.getFileDescriptor(file_path, "r")
        if not result["OK"]:
            result = fileHelper.sendEOF()
            # check if the file does not really exist
            if not os.path.exists(file_path):
                return S_ERROR("File %s does not exist" % os.path.basename(file_path))
            else:
                return S_ERROR("Failed to get file descriptor")

        fileDescriptor = result["Value"]
        result = fileHelper.FDToNetwork(fileDescriptor)
        if not result["OK"]:
            return S_ERROR("Failed to get file %s" % fileID)

        if os.path.exists(file_path):
            os.remove(file_path)

        return result

    def transfer_fromClient(self, fileID, token, fileSize, fileHelper):
        """Method to receive file from clients.
        fileID is the local file name in the SE.
        fileSize can be Xbytes or -1 if unknown.
        token is used for access rights confirmation.
        """
        if not self.__checkForDiskSpace(BASE_PATH, fileSize):
            return S_ERROR("Not enough disk space")

        file_path = "%s/%s" % (BASE_PATH, fileID)
        mkDir(os.path.dirname(file_path))
        result = fileHelper.getFileDescriptor(file_path, "w")
        if not result["OK"]:
            return S_ERROR("Failed to get file descriptor")

        fileDescriptor = result["Value"]
        result = fileHelper.networkToFD(fileDescriptor)
        if not result["OK"]:
            return S_ERROR("Failed to put file %s" % fileID)
        return result

    @staticmethod
    def __checkForDiskSpace(dpath, size):
        """Check if the directory dpath can accomodate 'size' volume of data"""
        dsize = (getDiskSpace(dpath) - 1) * 1024 * 1024
        maxStorageSizeBytes = 1024 * 1024 * 1024
        return min(dsize, maxStorageSizeBytes) > size



""" ProxyProvider base class for various proxy providers
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id$"


class ProxyProvider(object):
    def __init__(self, parameters=None):

        self.parameters = parameters
        self.name = None
        if parameters:
            self.name = parameters.get("ProviderName")

    def setParameters(self, parameters):
        self.parameters = parameters
        self.name = parameters.get("ProviderName")

    def checkStatus(self, userDN):
        """Read ready to work status of proxy provider

        :param str userDN: user DN

        :return: S_OK()/S_ERROR()
        """
        return S_OK()

    def generateDN(self, **kwargs):
        """Generate new DN

        :param dict kwargs: user description dictionary

        :return: S_OK(str)/S_ERROR() -- contain DN
        """
        return S_ERROR("Not implemented in %s", self.name)
